{
 "cells": [
  {
   "attachments": {},
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Chapter 4. Binary Search Tree\n",
    "Binary Search Tree (BST) is a node-based binary tree data structure which has the following properties: \n",
    "\n",
    "- The left subtree of a node contains only nodes with `keys less than the node’s key`.\n",
    "- The right subtree of a node contains only nodes with `keys greater than the node’s key`.\n",
    "- This means `everything to the left of the root is less than the value of the root and everything to the right of the root is greater than the value of the root`. Due to this performing, a binary search is very easy.\n",
    "- The left and right subtree each must also be a binary search tree.  \n",
    "- There must be no duplicate nodes (BST may have duplicate values with different handling approaches)\n",
    "\n",
    "**Terminology**\n",
    "- `traversal order`: the order of the nodes generated from a given traversal direction. \n",
    "- `preorder traversal`: repeatly traverse the node first, then the left and lastly the right. \n",
    "\n",
    "**Key Properties**\n",
    "- Keep track of the range of node value. -> can be used to check if a tree is a valid BST or construct a BST from given preorder traversal order.\n",
    "  "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [],
   "source": [
    "class Node():\n",
    "    def __init__(self, val=0, left=None, right=None):\n",
    "        self.val = val \n",
    "        self.left = left\n",
    "        self.right = right\n",
    "\n",
    "def traverse_preorder(root: Node) -> list:\n",
    "    if not root:\n",
    "        return []\n",
    "    \n",
    "    res = [root.val]\n",
    "    if root.left:\n",
    "        res.extend(traverse_preorder(root.left))\n",
    "    if root.right:\n",
    "        res.extend(traverse_preorder(root.right))\n",
    "    \n",
    "    return res\n",
    "\n",
    "def printTree(root, method='preorder'):\n",
    "    traversal = []\n",
    "    if method == \"preorder\":\n",
    "        pass"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Build\n",
    "\n",
    "### From Preorder\n",
    "\n",
    "How to build a BST from given preorder traversal? For example if the given preorder traversal is [10,5,1,7,40,50], then the BST has the following format:\n",
    "\n",
    "\n",
    "            10\n",
    "            / \\\n",
    "           5   40\n",
    "          / \\    \\\n",
    "         1   7   50\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "**Important**\n",
    "- In `preorder` traversal, the root node always comes first.\n",
    "- The key is to find its left subtree and right subtree from a given root using the above principle.\n",
    "- Together with `inorder`, where left subtree always comes first, it is easy to reconstruct a general binary tree.\n",
    "\n",
    "**Notes**\n",
    "- General BT construction needs both preorder and inorder traversals, but BST can be constructed from preorder traversal only or postorder traversal only. \n",
    "- BST cannot be constructed from inorder traveral only.\n",
    "\n",
    "**Leetcode**\n",
    "- [1008](https://leetcode.com/problems/construct-binary-search-tree-from-preorder-traversal/)"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "#### Method 1\n",
    "- The first element of the preorder traversal `A` is always root. \n",
    "- Then we find the first element that is greater than the root, and get its index `i`. In 0-index language, any values between `1` and `i` shall be the left subtree, and values between `i` and `-1` shall be the right subtree.\n",
    "- Recursively cconstruct the left subtree using `A[1:i]` and right subtree using `A[i:]`.\n",
    "\n",
    "**Complexity**\n",
    "- time: $O(n^2)$\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "[10, 5, 1, 7, 40, 50]\n",
      "[]\n",
      "[1]\n"
     ]
    }
   ],
   "source": [
    "def build_preorder(A):\n",
    "    # return None if empty\n",
    "    if not A:\n",
    "        return None\n",
    "\n",
    "    # get root\n",
    "    root = Node(val = A[0])\n",
    "    n = len(A)\n",
    "\n",
    "    # get left subtree and right subtree\n",
    "    # using the first element of the right subtree\n",
    "    # or we can use inorder traversal here by sorting the preorder traversal\n",
    "    i = 1\n",
    "    while i < n and A[i] < A[0]:\n",
    "        i += 1\n",
    "    \n",
    "    # recurse\n",
    "    left = build_preorder(A[1:i])\n",
    "    right = build_preorder(A[i:])\n",
    "    \n",
    "    # combine \n",
    "    root.left = left \n",
    "    root.right = right \n",
    "\n",
    "    return root\n",
    "    \n",
    "# test\n",
    "A = [10,5,1,7,40,50]\n",
    "root = build_preorder(A)\n",
    "print(traverse_preorder(root))\n",
    "\n",
    "A = []\n",
    "root = build_preorder(A)\n",
    "print(traverse_preorder(root))\n",
    "\n",
    "A = [1]\n",
    "root = build_preorder(A)\n",
    "print(traverse_preorder(root))"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "#### Method 2\n",
    "\n",
    "Traverse each element in the preorder traversal, and compare its value with the root value\n",
    "- if the node value is less than the root value, change root to its left\n",
    "- if the node value is greater than the root value, change root to its right\n",
    "- recurse until the node is inserted correctly\n",
    "\n",
    "**Complexity**\n",
    "- time: $O(n^2)$\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "[10, 5, 1, 7, 40, 50]\n",
      "[]\n",
      "[1]\n"
     ]
    }
   ],
   "source": [
    "def buildBST(preorder):\n",
    "    if not A:\n",
    "        return None \n",
    "    \n",
    "    root = Node(preorder[0])\n",
    "\n",
    "    n = len(preorder)\n",
    "    if n == 1:\n",
    "        return root\n",
    "    \n",
    "    # insert\n",
    "    def insert(root, node):\n",
    "        # base case\n",
    "        if not root:\n",
    "            return node\n",
    "        \n",
    "        # insert to left tree or right tree\n",
    "        if node.val < root.val:\n",
    "            root.left = insert(root.left, node)\n",
    "        else:\n",
    "            root.right = insert(root.right, node)\n",
    "        \n",
    "        return root\n",
    "\n",
    "    for i in range(1,n):\n",
    "        root = insert(root, Node(preorder[i]))\n",
    "    \n",
    "    return root\n",
    "\n",
    "# test\n",
    "A = [10,5,1,7,40,50]\n",
    "root = buildBST(A)\n",
    "print(traverse_preorder(root))\n",
    "\n",
    "A = []\n",
    "root = buildBST(A)\n",
    "print(traverse_preorder(root))\n",
    "\n",
    "A = [1]\n",
    "root = buildBST(A)\n",
    "print(traverse_preorder(root))"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "#### Method 3\n",
    "The information of the binary tree can be saved using preorder traversal and inorder traversal.\n",
    "\n",
    "The key idea here is:\n",
    "- preorder traversal preserves the node order in [root, left, right]\n",
    "- inorder traversal preserves the node order in [left, root, right]. For BST, the inorder traversal is sorted increasingly. \n",
    "  - we can get the inroder traversal of BST easily from sorting the preorder traversal.\n",
    "  - The first element in the preorder traversal is always the root, and the left slice of the inorder traversal to the root is the whole left subtree.\n",
    "  - Therefore, we just need to know the index of the root in the inorder traversal. \n",
    "\n",
    "**Complexity**\n",
    "- time: $O(n^2log(n))$\n",
    "- space: $O(n)$. Extra space for storing inorder traversal and recursive call stack, whose worst scenario is $O(n)$ when the tree is left-skewed.\n",
    "\n",
    "**This could be improved to $O(n)$ time**"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "[10, 5, 1, 7, 40, 50]\n",
      "[10, 5, 3, 1, 4, 7, 6, 8, 40, 50]\n"
     ]
    }
   ],
   "source": [
    "def buildBST(preorder):\n",
    "    # base case\n",
    "    if len(preorder) < 1:\n",
    "        return None\n",
    "    \n",
    "    root = Node(preorder[0])\n",
    "    # get inorder traversal of a BST\n",
    "    inorder = sorted(preorder)\n",
    "\n",
    "    # find root index in inorder traversal\n",
    "    ind = inorder.index(preorder[0])\n",
    "    \n",
    "    # build left\n",
    "    root.left = buildBST(preorder[1:ind+1])\n",
    "    root.right = buildBST(preorder[ind+1:])\n",
    "\n",
    "    return root\n",
    "\n",
    "# test\n",
    "A = [10,5,1,7,40,50]\n",
    "root = buildBST(A)\n",
    "print(traverse_preorder(root))\n",
    "\n",
    "A = [10, 5, 3, 1, 4, 7, 6, 8, 40, 50]\n",
    "root = buildBST(A)\n",
    "print(traverse_preorder(root))"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "#### Method 4\n",
    "See https://www.youtube.com/watch?v=UmJT3j26t1I.\n",
    " \n",
    "BST node has a property that determines the range of its subtree.\n",
    "The key idea here is use the upper bounds of each explored node to decide where a specific node should be inserted:\n",
    "- initialize the upper bound of the root as `INF`\n",
    "- for each node, recursively\n",
    "  - build the left subtree using upper bound `node.val`\n",
    "  - build the right subtree using upper bound the current node's parent upper bound\n",
    "\n",
    "**Complexity**\n",
    "- time: O(n). For each node, we at most visit 3 times, thus the worst scenario is $O(3n)$, which is $O(n)$\n",
    "- space: O(n). This recursive algorithm uses stack space. \n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "A for left: [5, 1, 7, 40, 50]\n",
      "A for left: [1, 7, 40, 50]\n",
      "A for left: [7, 40, 50]\n",
      "A for right: [7, 40, 50]\n",
      "A for right: [7, 40, 50]\n",
      "A for left: [40, 50]\n",
      "A for right: [40, 50]\n",
      "A for right: [40, 50]\n",
      "A for left: [50]\n",
      "A for right: [50]\n",
      "A for left: []\n",
      "A for right: []\n",
      "[10, 5, 1, 7, 40, 50]\n"
     ]
    }
   ],
   "source": [
    "# The recursion is so beatiful here. Try to understand\n",
    "# The key point here is: \n",
    "# 1. A is globally changed by each recursion.\n",
    "# 2. The recusion always build left brach first. Until no left node is possible, it will build the right branch from the bottom of left subtree to the right subtree of the root.\n",
    "def buildBST(preorder):\n",
    "    def build(A, ub=float('inf')):\n",
    "        if not A or A[0] > ub: return None\n",
    "        #print(A)\n",
    "        root = Node(A.pop(0))\n",
    "\n",
    "        print(f\"A for left: {A}\")\n",
    "        root.left = build(A, root.val)\n",
    "        print(f\"A for right: {A}\")\n",
    "        root.right = build(A, ub)\n",
    "        \n",
    "        return root\n",
    "\n",
    "    return build(preorder) \n",
    "\n",
    "# test\n",
    "A = [10,5,1,7,40,50]\n",
    "root = buildBST(A)\n",
    "print(traverse_preorder(root))\n",
    "\n",
    "#A = []\n",
    "#root = buildBST(A)\n",
    "#print(traverse_preorder(root))\n",
    "\n",
    "#A = [1]\n",
    "#root = buildBST(A)\n",
    "#print(traverse_preorder(root))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "[10, 5, 1, 7, 40, 50]\n"
     ]
    }
   ],
   "source": [
    "# A wrong implementation:\n",
    "# the index i is not global, which results the same node will be attached to both left and right if possible.\n",
    "def buildBST(preorder):\n",
    "    \n",
    "    def build(A, i, ub=float('inf')):\n",
    "        if i == len(A) or A[i] > ub:\n",
    "            return None\n",
    "        root = Node(A[i])\n",
    "        root.left = build(A, i+1, root.val)\n",
    "        root.right = build(A, i+1, ub)\n",
    "        \n",
    "        return root\n",
    "\n",
    "    return build(preorder, 0) \n",
    "\n",
    "# The following implementation fixes the above code\n",
    "def buildBST(preorder):\n",
    "    global i \n",
    "    i = 0\n",
    "    def build(A, ub=float('inf')):\n",
    "        global i\n",
    "        if i == len(A) or A[i] > ub:\n",
    "            return None\n",
    "        root = Node(A[i])\n",
    "        i += 1\n",
    "        root.left = build(A, root.val)\n",
    "        root.right = build(A, ub)\n",
    "        \n",
    "        return root\n",
    "\n",
    "    return build(preorder) \n",
    "\n",
    "# test\n",
    "A = [10,5,1,7,40,50]\n",
    "root = buildBST(A)\n",
    "print(traverse_preorder(root))"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Build All BST from Inorder Traversal\n",
    "\n",
    "Inorder traversal alone cannot define a unique BST. But based on inorder traversal, we can get all the possible BST structures. How?\n",
    "\n",
    "See [here](https://labuladong.github.io/algo/di-yi-zhan-da78c/shou-ba-sh-66994/dong-ge-da-b16de/).\n",
    "\n",
    "This will require dynamic programming. We will come back later."
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Traverse \n",
    "\n",
    "Traverse the BST node by node.\n",
    "\n",
    "- `Depth-first Search`\n",
    "  - preorder\n",
    "  - inorder\n",
    "  - postorder\n",
    "- `Breadth-first Search`\n",
    "  - level-order \n",
    "\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Preorder Traversal\n",
    "Preorder traversal is used to create a copy of the tree. Preorder traversal is also used to get prefix expressions on an expression tree.\n",
    "- visit the root\n",
    "- traverse the left subtree\n",
    "- traverse the right subtree"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 31,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "[10, 5, 1, 7, 40, 50]\n"
     ]
    }
   ],
   "source": [
    "# Method 1: dynamic programming: divide into subproblem and integrate subproblem solutions to get final solution\n",
    "def preorder(root):\n",
    "    # base case\n",
    "    if not root:\n",
    "        return []\n",
    "\n",
    "    # store node \n",
    "    order = [root.val]\n",
    "    # traverse the left subtree and right subtree\n",
    "    order.extend(preorder(root.left))\n",
    "    order.extend(preorder(root.right))\n",
    "    \n",
    "    # return\n",
    "    return order\n",
    "\n",
    "A = [10,5,1,7,40,50]\n",
    "root = buildBST(A)\n",
    "print(preorder(root))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 34,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "[10, 5, 1, 7, 40, 50]\n"
     ]
    }
   ],
   "source": [
    "# Method 2: backtracking \n",
    "def preorder(root):\n",
    "\n",
    "    # traverse function\n",
    "    def traverse(root):\n",
    "        \n",
    "        if not root:\n",
    "            return \n",
    "\n",
    "        res.append(root.val)\n",
    "        traverse(root.left)\n",
    "        traverse(root.right)\n",
    "\n",
    "    res = []\n",
    "    traverse(root)\n",
    "    return res\n",
    "\n",
    "A = [10,5,1,7,40,50]\n",
    "root = buildBST(A)\n",
    "print(preorder(root))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Inorder\n",
    "In the case of BST, Inorder traversal gives nodes in non-decreasing order. To get nodes of BST in non-increasing order, a variation of Inorder traversal where Inorder traversal is reversed can be used. \n",
    "- traverse the left subtree\n",
    "- visit the root\n",
    "- traverse the rigth subtree\n",
    "\n",
    "\n",
    "**Recursive Implementation**"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "[1, 5, 7, 10, 40, 50]\n"
     ]
    }
   ],
   "source": [
    "## A recursive implementation\n",
    "def inorder(root):\n",
    "    # base case\n",
    "    if not root:\n",
    "        return []\n",
    "\n",
    "    order = []\n",
    "    # traverse left subtree\n",
    "    order.extend(inorder(root.left))\n",
    "    order.extend([root.val])\n",
    "    order.extend(inorder(root.right))\n",
    "\n",
    "    return order\n",
    "\n",
    "A = [10,5,1,7,40,50]\n",
    "root = buildBST(A)\n",
    "print(inorder(root))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "**Iterative Implementation Using Stack**\n",
    "1. Create an empty stack S.\n",
    "2. Initialize current node as root\n",
    "3. Push the current node to S and set `current = current.left` until current is NULL\n",
    "4. If current is NULL and stack is not empty then \n",
    "     - Pop the top item from stack.\n",
    "     - Print the popped item, set `current = popped_item.right` \n",
    "     - Go to step 3.\n",
    "5. If current is NULL and stack is empty then we are done\n",
    "\n",
    "     ```shell\n",
    "     ============================================================\n",
    "     Let us consider the below tree for example  \n",
    "\n",
    "     \n",
    "                 1\n",
    "               /   \\\n",
    "             2      3\n",
    "            /  \\\n",
    "          4     5\n",
    "\n",
    "     Step 1 Creates an empty stack: S = NULL\n",
    "\n",
    "     Step 2 sets current as address of root: current -> 1\n",
    "\n",
    "     Step 3 Pushes the current node and set current = current->left \n",
    "          until current is NULL\n",
    "          current -> 1\n",
    "          push 1: Stack S -> 1\n",
    "          current -> 2\n",
    "          push 2: Stack S -> 2, 1\n",
    "          current -> 4\n",
    "          push 4: Stack S -> 4, 2, 1\n",
    "          current = NULL\n",
    "\n",
    "     Step 4 pops from S\n",
    "          a) Pop 4: Stack S -> 2, 1\n",
    "          b) print \"4\"\n",
    "          c) current = NULL /*right of 4 */ and go to step 3\n",
    "     Since current is NULL step 3 doesn't do anything. \n",
    "\n",
    "     Step 4 pops again.\n",
    "          a) Pop 2: Stack S -> 1\n",
    "          b) print \"2\"\n",
    "          c) current -> 5/*right of 2 */ and go to step 3\n",
    "\n",
    "     Step 3 pushes 5 to stack and makes current NULL\n",
    "          Stack S -> 5, 1\n",
    "          current = NULL\n",
    "\n",
    "     Step 4 pops from S\n",
    "          a) Pop 5: Stack S -> 1\n",
    "          b) print \"5\"\n",
    "          c) current = NULL /*right of 5 */ and go to step 3\n",
    "     Since current is NULL step 3 doesn't do anything\n",
    "\n",
    "     Step 4 pops again.\n",
    "          a) Pop 1: Stack S -> NULL\n",
    "          b) print \"1\"\n",
    "          c) current -> 3 /*right of 1 */  \n",
    "\n",
    "     Step 3 pushes 3 to stack and makes current NULL\n",
    "          Stack S -> 3\n",
    "          current = NULL\n",
    "\n",
    "     Step 4 pops from S\n",
    "          a) Pop 3: Stack S -> NULL\n",
    "          b) print \"3\"\n",
    "          c) current = NULL /*right of 3 */  \n",
    "\n",
    "     Traversal is done now as stack S is empty and current is NULL.\n",
    "     ``` "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 10,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "[1, 5, 7, 10, 40]\n"
     ]
    }
   ],
   "source": [
    "## A iterative implementation: use a stack to store \n",
    "def inorder(root):\n",
    "    order = []\n",
    "    stack = [root] # if use stack=[root]\n",
    "    curr = root\n",
    "    # Note len([None]=1)\n",
    "    while stack:\n",
    "        # pop if current is none and stack is not empty\n",
    "        if not curr:\n",
    "            curr = stack.pop()\n",
    "            # get the order\n",
    "            order.append(curr.val)\n",
    "            # update current node\n",
    "            curr = curr.right\n",
    "        \n",
    "        # push all left nodes and then right node\n",
    "        while curr:\n",
    "            # avoid storing root twice\n",
    "            if curr.val != root.val:\n",
    "                stack.append(curr)\n",
    "            curr = curr.left\n",
    "\n",
    "    return order\n",
    "\n",
    "# Test\n",
    "A = [10,5,1,7,40]\n",
    "root = buildBST(A)\n",
    "print(inorder(root))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "**Iterative Implementation without Stack**\n",
    "\n",
    "This method is known as Morris Traversal.\n",
    "See [here](https://www.geeksforgeeks.org/inorder-tree-traversal-without-recursion-and-without-stack/?ref=lbp)."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Postorder\n",
    "Postorder traversal is used to delete the tree. Please see the question for the deletion of a tree for details. Postorder traversal is also useful to get the postfix expression of an expression tree\n",
    "- traverse the left subtree\n",
    "- traverse the right subtree\n",
    "- visit the root\n",
    "  "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 11,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "[1, 5, 7, 10, 40, 50]\n",
      "[1, 7, 5, 50, 40, 10]\n"
     ]
    }
   ],
   "source": [
    "def postorder(root):\n",
    "    # base \n",
    "    if not root:\n",
    "        return []\n",
    "    \n",
    "    order = []\n",
    "    # traverse \n",
    "    order.extend(postorder(root.left))\n",
    "    order.extend(postorder(root.right))\n",
    "    order.extend([root.val])\n",
    "\n",
    "    # return\n",
    "    return order\n",
    "\n",
    "A = [10,5,1,7, 40, 50]\n",
    "root = buildBST(A)\n",
    "print(inorder(root))\n",
    "print(postorder(root))"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Level Order Traversal\n",
    "\n",
    "Traverse the tree level by level from top to bottom\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 23,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "[1, 5, 7, 10, 40, 50]\n",
      "[[10], [5, 40], [1, 7, 'null', 50]]\n"
     ]
    }
   ],
   "source": [
    "def levelOrderTop(root):\n",
    "    if not root:\n",
    "        return []\n",
    "    \n",
    "    # initialize\n",
    "    level = 1\n",
    "    queue = [(root, level)]\n",
    "    res = []\n",
    "\n",
    "    while queue:\n",
    "        level_res = []\n",
    "        level = queue[0][1]\n",
    "\n",
    "        # traverse each level\n",
    "        while queue and level == queue[0][1]:\n",
    "            # pop all nodes from the same level\n",
    "            node, level = queue.pop(0)\n",
    "\n",
    "            # stack level result\n",
    "            if node:\n",
    "                level_res.append(node.val)\n",
    "            else:\n",
    "                level_res.append('null')\n",
    "            \n",
    "            # update queue: node at least has one child\n",
    "            if node and (node.left or node.right):\n",
    "                queue.extend([(node.left, level+1), (node.right, level+1)])\n",
    "\n",
    "        # stack level results to head:\n",
    "        res.append(level_res)\n",
    "    \n",
    "    return res\n",
    "\n",
    "A = [10,5,1,7, 40, 50]\n",
    "root = buildBST(A)\n",
    "print(inorder(root))\n",
    "print(levelOrderTop(root))"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Traverse the tree level by level from bottom to top\n",
    "\n",
    "*Method 1:*\n",
    "- use a stack to store all nodes and level info level-by-level from top to bottom\n",
    "- pop the stack to another stack so that all the nodes in the same level are in the new stack\n",
    "  "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 27,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "[[1, 7, 'null', 50], [5, 40], [10]]\n"
     ]
    }
   ],
   "source": [
    "def levelOrderBottom(root):\n",
    "    if not root:\n",
    "            return []\n",
    "\n",
    "    level = 1\n",
    "    queue = [(root, level)]\n",
    "\n",
    "    # initialize two stacks\n",
    "    stack = []\n",
    "\n",
    "    # initialize output\n",
    "    res= []\n",
    "\n",
    "    # traverse to store all nodes in stack\n",
    "    while queue:\n",
    "        node, level = queue.pop(0)\n",
    "\n",
    "        # save to a stack\n",
    "        stack.append((node, level))\n",
    "\n",
    "        # append children to queue: keep null as a place holder\n",
    "        if node and (node.left or node.right):\n",
    "            queue.extend([(node.left, level + 1), (node.right, level + 1)])\n",
    "\n",
    "    # pop stack to get level order traversal from bottom\n",
    "    while stack:\n",
    "        level_stack = []\n",
    "        # traverse level\n",
    "        level = stack[-1][1]\n",
    "        while stack and level == stack[-1][1]:\n",
    "            node, level = stack.pop()\n",
    "            level_stack.append(node)\n",
    "        \n",
    "        # pop level_stack to get the output\n",
    "        level_res = []\n",
    "        while level_stack:\n",
    "            node = level_stack.pop()\n",
    "            if node:\n",
    "                level_res.append(node.val)\n",
    "            else:\n",
    "                level_res.append('null')    \n",
    "        # output \n",
    "        res.append(level_res)\n",
    "    \n",
    "    return res\n",
    "\n",
    "A = [10,5,1,7, 40, 50]\n",
    "root = buildBST(A)\n",
    "print(levelOrderBottom(root))"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "*Method 2:*\n",
    "\n",
    "This method is the same as level traversal from top to down, but when appending level results, we insert the level result at the head instead of tail."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 20,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "[[1, 7, 'null', 50], [5, 40], [10]]\n"
     ]
    }
   ],
   "source": [
    "def levelOrderBottom(root):\n",
    "    if not root:\n",
    "        return []\n",
    "    \n",
    "    # initialize\n",
    "    level = 1\n",
    "    queue = [(root, level)]\n",
    "    res = []\n",
    "\n",
    "    while queue:\n",
    "        level_res = []\n",
    "        level = queue[0][1]\n",
    "\n",
    "        # traverse each level\n",
    "        while queue and level == queue[0][1]:\n",
    "            # pop all nodes from the same level\n",
    "            node, level = queue.pop(0)\n",
    "\n",
    "            # stack level result\n",
    "            if node:\n",
    "                level_res.append(node.val)\n",
    "            else:\n",
    "                level_res.append('null')\n",
    "            \n",
    "            # update queue: node at least has one child\n",
    "            if node and (node.left or node.right):\n",
    "                queue.extend([(node.left, level+1), (node.right, level+1)])\n",
    "\n",
    "        # stack level results to head:\n",
    "        res.insert(0, level_res)\n",
    "    \n",
    "    return res\n",
    "\n",
    "A = [10,5,1,7, 40, 50]\n",
    "root = buildBST(A)\n",
    "print(levelOrderBottom(root))"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Serialization and Deserialization\n",
    "\n",
    "Serialization is a way to convert a binary tree to a standard format, which then can be used by external tools.\n",
    "-\n",
    "\n",
    "Deserialization is a way to construct a binary tree from a given serialization format.\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 29,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "[10, 5, 40, 1, 7, 'null', 50]\n"
     ]
    }
   ],
   "source": [
    "def serialize(root, order = \"level-order\"):\n",
    "    \n",
    "    def level_order_serialize(root):\n",
    "        if not root:\n",
    "            return []\n",
    "        \n",
    "        # traverse level by level from top to down, and use null as a placeholder for missing child or children\n",
    "        queue = [root]\n",
    "        res = []\n",
    "        while queue:\n",
    "            node = queue.pop(0)\n",
    "            val = node.val if node else \"null\"\n",
    "            res.append(val)\n",
    "\n",
    "            # update queue\n",
    "            if node and (node.left or node.right):\n",
    "                queue.extend([node.left, node.right])\n",
    "\n",
    "        return res\n",
    "\n",
    "    if order == \"level-order\":\n",
    "        return level_order_serialize(root)\n",
    "    else:\n",
    "        pass\n",
    "\n",
    "A = [10,5,1,7, 40, 50]\n",
    "root = buildBST(A)\n",
    "print(serialize(root, \"level-order\"))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "def deserialize(serialization, order='level-order'):\n",
    "    # serialization represent a binary tree in its complete format by introducing null whereever possible.\n",
    "    if not serialization:\n",
    "        return None\n",
    "    \n",
    "    def level_order_deserialize(serialization):\n",
    "        pass\n",
    "       "
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Insertion\n",
    "\n",
    "A new key is always inserted at the leaf. We start searching for a key from the root until we hit a leaf node. Once a leaf node is found, the new node is added as a child of the leaf node.\n",
    "\n",
    "```text\n",
    "        Insert 3 to the following BST will lead to:\n",
    "\n",
    "             10                                              10\n",
    "            /  \\                     insert 3               /  \\\n",
    "           5   40                    ------->              5   40              \n",
    "          / \\    \\                                        / \\    \\\n",
    "         1  7    50                                      1   7   50\n",
    "                                                          \\\n",
    "                                                           3\n",
    "```"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "[1, 3, 5, 7, 10, 40, 50]\n",
      "[10, 5, 1, 3, 7, 40, 50]\n"
     ]
    }
   ],
   "source": [
    "# A recursion Implementation\n",
    "def insert(root, key):\n",
    "    node = Node(key)\n",
    "    # base case\n",
    "    if not root:\n",
    "        return node\n",
    "    \n",
    "    # insert recursively assuming no duplicates\n",
    "    if root.val > key:\n",
    "        root.left = insert(root.left, key)\n",
    "    else:\n",
    "        root.right = insert(root.right, key)\n",
    "    \n",
    "    return root\n",
    "\n",
    "A = [10,5,1,7, 40, 50]\n",
    "root = buildBST(A)\n",
    "root = insert(root, 3)\n",
    "print(inorder(root))    \n",
    "print(preorder(root))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "[1, 3, 5, 7, 10, 40, 50]\n",
      "[10, 5, 1, 3, 7, 40, 50]\n"
     ]
    }
   ],
   "source": [
    "# An iterative implementation\n",
    "def insert(root, key):\n",
    "    node = Node(key)\n",
    "    cur = root \n",
    "    prev = None \n",
    "\n",
    "    while cur:\n",
    "        prev = cur\n",
    "        # go left\n",
    "        if cur.val > key:\n",
    "            cur = cur.left\n",
    "        else: # go right\n",
    "            cur = cur.right\n",
    "    \n",
    "    # specify connections\n",
    "    if prev.val > key:\n",
    "        prev.left = node \n",
    "    else:\n",
    "        prev.right = node\n",
    "    \n",
    "    return root\n",
    "\n",
    "A = [10,5,1,7, 40, 50]\n",
    "root = buildBST(A)\n",
    "root = insert(root, 3)\n",
    "print(inorder(root))    \n",
    "print(preorder(root))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Deletion\n",
    "When we delete a node, three possibilities arise. \n",
    "\n",
    "1) Node to be deleted is the leaf: Simply remove from the tree. \n",
    "\n",
    "```text\n",
    "              50                            50\n",
    "           /     \\         delete(20)      /   \\\n",
    "          30      70       --------->    30     70 \n",
    "         /  \\    /  \\                     \\    /  \\ \n",
    "       20   40  60   80                   40  60   80\n",
    "```\n",
    "\n",
    "2) Node to be deleted has only one child: Copy the child to the node and delete the child \n",
    "\n",
    "```text\n",
    "              50                            50\n",
    "           /     \\         delete(30)      /   \\\n",
    "          30      70       --------->    40     70 \n",
    "            \\    /  \\                          /  \\ \n",
    "            40  60   80                       60   80\n",
    "```\n",
    "3) Node to be deleted has two children: Find `inorder successor` of the node. Copy contents of the inorder successor to the node and delete the inorder successor. Note that `inorder predecessor` can also be used. \n",
    "\n",
    "```text\n",
    "              50                            60\n",
    "           /     \\         delete(50)      /   \\\n",
    "          40      70       --------->    40    70 \n",
    "                 /  \\                            \\ \n",
    "                60   80                           80\n",
    "```\n",
    "The important thing to note is, inorder successor is needed only when the right child is not empty. In this particular case, inorder successor can be obtained by finding the minimum value in the right child of the node.\n",
    "\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "[50, 30, 40, 70, 60, 80]\n",
      "[50, 40, 70, 60, 80]\n",
      "[60, 40, 70, 80]\n"
     ]
    }
   ],
   "source": [
    "def delete(root, key):\n",
    "\n",
    "    def successor(root):\n",
    "        \"\"\"find the inorder successor for a given node\n",
    "            - if root has no right child, return None\n",
    "            - else return the leftmost node in the right subtree\n",
    "        \"\"\"\n",
    "        if not root or not root.right:\n",
    "            return None \n",
    "        \n",
    "        node = root.right\n",
    "        while node.left:\n",
    "            node = node.left\n",
    "\n",
    "        return node\n",
    "\n",
    "    # base case\n",
    "    if not root:\n",
    "        return root \n",
    "    \n",
    "    # if key is smaller than the root's key then it lies in the left subtree\n",
    "    if key < root.val:\n",
    "        root.left = delete(root.left, key)\n",
    "    elif key > root.val:\n",
    "        root.right = delete(root.right, key)\n",
    "    # key is the same as the root key\n",
    "    else:\n",
    "        # root is a leaf\n",
    "        if not root.left and not root.right:\n",
    "            return None\n",
    "        # root has only one child - left \n",
    "        elif root.left and not root.right:\n",
    "            temp = root.left\n",
    "            root = None \n",
    "            return temp\n",
    "        # root has only one child - right\n",
    "        elif not root.left and root.right:\n",
    "            temp = root.right\n",
    "            root = None \n",
    "            return temp \n",
    "        \n",
    "        # root has two children\n",
    "        succ = successor(root)\n",
    "\n",
    "        # copy the content\n",
    "        #right = root.right # need reverse the right subtree before deletion\n",
    "        #left = root.left\n",
    "        root.val = succ.val\n",
    "\n",
    "        # delete the inorder successor\n",
    "        #root.left = left\n",
    "        root.right = delete(root.right, succ.val)\n",
    "\n",
    "    return root\n",
    "\n",
    "# test\n",
    "A = [50,30,20,40,70,60,80]\n",
    "root = buildBST(A)  \n",
    "root = delete(root, 20)\n",
    "print(preorder(root))\n",
    "\n",
    "A = [50,30,40,70,60,80]\n",
    "root = buildBST(A)  \n",
    "root = delete(root, 30)\n",
    "print(preorder(root))\n",
    "\n",
    "A = [50,40,70,60,80]\n",
    "root = buildBST(A)  \n",
    "root = delete(root, 50)\n",
    "print(preorder(root))"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Search\n",
    "Search in binary tree is very simple which is the same as binary search.\n",
    "\n",
    "### Search an element\n",
    "\n",
    "```\n",
    "ITERATIVE_TREE_SEARCH(root, x):\n",
    "    while x and x != root.val\n",
    "        if x < root.key:\n",
    "            root = root.left\n",
    "        else\n",
    "            root = root.right\n",
    "        return root\n",
    "```\n",
    "\n",
    "### Minimum and Maximum\n",
    "- `Minimum`: go to leftmost leaf node\n",
    "- `Maximum`: go to rightmost leaf node\n",
    "\n",
    "### Successor and Predecessor\n",
    "- `Successor`: inorder sucessor is the next element in the inorder traversal order. \n",
    "  - If current node has right subtree, then return left most leaf of the right-subtree (for BST, which is the minimum node of the right subtree)\n",
    "  - If current node has no right subtree, then return the lowest ancestor"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Applications\n",
    "\n",
    "### Binary Tree to BST \n",
    "Given a Binary Tree, convert it to a Binary Search Tree. The conversion must be done in such a way that keeps the original structure of Binary Tree.\n",
    "\n",
    "```\n",
    "Example 1\n",
    "Input:\n",
    "          10\n",
    "         /  \\\n",
    "        2    7\n",
    "       / \\\n",
    "      8   4\n",
    "Output:\n",
    "          8\n",
    "         /  \\\n",
    "        4    10\n",
    "       / \\\n",
    "      2   7\n",
    "\n",
    "\n",
    "Example 2\n",
    "Input:\n",
    "          10\n",
    "         /  \\\n",
    "        30   15\n",
    "       /      \\\n",
    "      20       5\n",
    "Output:\n",
    "          15\n",
    "         /  \\\n",
    "       10    20\n",
    "       /      \\\n",
    "      5        30\n",
    "```\n",
    "\n",
    "### Analysis\n",
    "three steps:\n",
    "- get inorder traveral as an array from binary tree\n",
    "- sort the array for BST\n",
    "- inorder traverse the binary tree, and copy to the value of each node from the sorted array. We have to do inorder, because the sorted array is an inorder representation of BST.\n",
    "\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "[10, 30, 20, 15, 5]\n",
      "[15, 10, 5, 20, 30]\n"
     ]
    }
   ],
   "source": [
    "def convertBTtoBST(root):\n",
    "\n",
    "    def traverse_inorder(root):\n",
    "        if not root:\n",
    "            return []\n",
    "        \n",
    "        l = traverse_inorder(root.left)\n",
    "        v = [root.val]\n",
    "        r = traverse_inorder(root.right)\n",
    "\n",
    "        return l + v + r\n",
    "    \n",
    "    def buildBST(root, arr_sort):\n",
    "        if root is None:\n",
    "            return \n",
    "        # no preorder operation\n",
    "\n",
    "        # traverse left subtree\n",
    "        left = buildBST(root.left, arr_sort)\n",
    "\n",
    "        # inorder operation - replace value\n",
    "        v = arr_sort.pop(0)\n",
    "        root.val = v\n",
    "\n",
    "        # traverse right subtree\n",
    "        right = buildBST(root.right, arr_sort)\n",
    "        # no postorder operation\n",
    "\n",
    "        return root\n",
    "\n",
    "    arr = traverse_inorder(root)\n",
    "    arr_sort = sorted(arr)\n",
    "\n",
    "    return buildBST(root, arr_sort)\n",
    "\n",
    "# test\n",
    "root = Node(10)\n",
    "root.left = Node(30)\n",
    "root.right = Node(15)\n",
    "root.left.left = Node(20)\n",
    "root.right.right = Node(5)\n",
    "print(traverse_preorder(root))\n",
    "bst = convertBTtoBST(root)\n",
    "print(traverse_preorder(bst))\n"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Tree Map\n",
    "\n",
    "See Leetcode 1038 and 538.\n",
    "\n",
    "The problem is to map a BST to a greater tree or a geater sum tree.\n",
    "\n",
    "For a greater tree, convert a BST to a Greater Tree such that every key of the original BST is changed to the original key plus the sum of all keys greater than the original key in BST.\n",
    "\n",
    "### Analysis\n",
    "- inorder traversal of a BST is a sequence with increasingly sorted keys. One naive apporach is:\n",
    "    - get inorder traversal of BST\n",
    "    - get greater sum of the inorder sequences\n",
    "    - build binary tree with the greater sum sequence but follow the original BST structure\n",
    "    - this method is O(n) in both time and complexity\n",
    "- How can we implement it using sub-problem method?\n",
    "    - maintain a sum that recursively add the current value to previous greater sum \n",
    "    - inorder traverse the BST with decreasing keys (traverse right then left)\n",
    "    - inorder change the node value with presum\n"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Valid BST\n",
    "It looks easy to check if a binary tree is a valid BST by simply comparing the keys in node, node.left, and node.right. \n",
    "\n",
    "But there is a trap using this method. Look at the below tree, where the left subtree and right subtree are both valid BST, and the node value 5<10<15, but it is not a valid BST as a whole.\n",
    "\n",
    "         10\n",
    "        /  \\\n",
    "       5    15\n",
    "           /  \\\n",
    "          6   20\n",
    "\n",
    "The reason is the above method doesn't guarantee that all values in the right subtree is greater than the node value, and all values in the left subtree should be smaller than the node value.\n",
    "\n",
    "For binary tree, we know the upper and lower bound for each node. We can use this property to check if all nodes are within correct bounds.\n",
    "\n",
    "**Pesudo code**:\n",
    "\n",
    "```\n",
    "boolean isValidBST(TreeNode root) {\n",
    "    return isValidBST(root, null, null);\n",
    "}\n",
    "\n",
    "/* 限定以 root 为根的子树节点必须满足 max.val > root.val > min.val */\n",
    "boolean isValidBST(TreeNode root, TreeNode min, TreeNode max) {\n",
    "    // base case\n",
    "    if (root == null) return true;\n",
    "    // 若 root.val 不符合 max 和 min 的限制，说明不是合法 BST\n",
    "    if (min != null && root.val <= min.val) return false;\n",
    "    if (max != null && root.val >= max.val) return false;\n",
    "    // 限定左子树的最大值是 root.val，右子树的最小值是 root.val\n",
    "    return isValidBST(root.left, min, root) \n",
    "        && isValidBST(root.right, root, max);\n",
    "}\n",
    "```"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# AVL Tree\n",
    "AVL tree is a self-balancing Binary Search Tree (BST) where the difference between heights of left and right subtrees cannot be more than one for all nodes. \n",
    "\n",
    "Most of the BST operations (e.g., search, max, min, insert, delete.. etc) take O(h) time where h is the height of the BST. The cost of these operations may become O(n) for a skewed Binary tree. If we make sure that the height of the tree remains O(log(n)) after every insertion and deletion, then we can guarantee an upper bound of O(log(n)) for all these operations. The height of an AVL tree is always O(log(n)) where n is the number of nodes in the tree.\n",
    "\n",
    "## Insertion\n",
    "To make sure that the given tree remains AVL after every insertion, we must augment the standard BST insert operation to perform some re-balancing without violating BST properties.\n",
    "Two operations can be performed:\n",
    "- left rotation\n",
    "- right rotation\n",
    "\n",
    "\n",
    "** Steps for Insertion **\n",
    "\n",
    "Let the newly inserted node be `w` \n",
    "\n",
    "- Perform standard BST insert for `w`. \n",
    "- Starting from `w`, travel up and find the first unbalanced node. Let `z` be the first unbalanced node, `y` be the child of `z` that comes on the path from `w` to `z` and `x` be the grandchild of `z` that comes on the path from `w` to `z`. \n",
    "- Re-balance the tree by performing appropriate rotations on the subtree rooted with `z`. There can be 4 possible cases that need to be handled as `x`, `y` and `z` can be arranged in 4 ways.\n",
    "- Following are the possible 4 arrangements: \n",
    "    - `y` is the left child of `z` and `x` is the left child of `y` (Left Left Case) \n",
    "    - `y` is the left child of `z` and `x` is the right child of `y` (Left Right Case) \n",
    "    - `y` is the right child of `z` and `x` is the right child of `y` (Right Right Case) \n",
    "    - `y` is the right child of `z` and `x` is the left child of `y` (Right Left Case)\n",
    "\n"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Red-Black Tree\n",
    "\n",
    "red-black tree is a binary search tree with one extra bit of storage per node: its color, which can be either red or black. \n",
    "1. Every node is either red or black.\n",
    "2. The root is black.\n",
    "3. Every leaf is black.\n",
    "4. If a node is red, then both its children are black.\n",
    "5. For each node, all simple paths from the node to descendant leaves contain the same number of black nodes."
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "jupyterbook-py38",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.8.12"
  },
  "orig_nbformat": 4,
  "vscode": {
   "interpreter": {
    "hash": "577791f2da4ac3a13c44f20f0f30021788d0264a6d82c3224a80270143730ffe"
   }
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
